from os.path import join

import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import FloatParam, IntParam


class Bioelements(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {"elements": DataType.TABLE}

    def config_params():
        return {
            "bioelements": {
                "tolerance": FloatParam(0.5, 0.01, 5),
                "n_elements": IntParam(2, 1, 4),
            }
        }

    def setup(self):
        import pandas as pd

        # load the dataframe here to avoid loading it on startup
        self.air_elements = pd.read_csv(join(self.assets_path, "air_elements_filtered.csv"))

    def process(self, data: Data):
        if data is None:
            return None

        data.data = np.squeeze(data.data)
        if data.data.ndim > 1:
            raise ValueError("Data must be 1D")

        tolerance = self.params["bioelements"]["tolerance"].value
        elems, spec_regions, types = bioelements_realtime(data.data, self.air_elements, tolerance)

        # select the n most common elements
        elems = elems[: self.params["bioelements"]["n_elements"].value]
        spec_regions = spec_regions[: self.params["bioelements"]["n_elements"].value]
        types = types[: self.params["bioelements"]["n_elements"].value]

        # remove duplicates
        elems = list(dict.fromkeys(elems))
        spec_regions = list(dict.fromkeys(spec_regions))
        types = list(dict.fromkeys(types))
        # create single string for each variable
        elems = ", ".join(elems)
        spec_regions = ", ".join(spec_regions)
        types = ", ".join(types)
        # combine the three lists into a dictionary
        elements = {
            "element": Data(DataType.STRING, elems, {}),
            "spectral_region": Data(DataType.STRING, spec_regions, {}),
            "type": Data(DataType.STRING, types, {}),
        }
        # print(elements)
        return {"elements": (elements, data.meta)}


hertz_to_nm_fn, find_matching_spectral_lines_fn = None, None


def bioelements_realtime(data, df, tolerance):
    global hertz_to_nm_fn, find_matching_spectral_lines_fn
    if hertz_to_nm_fn is None or find_matching_spectral_lines_fn is None:
        from biotuner.bioelements import find_matching_spectral_lines, hertz_to_nm

        hertz_to_nm_fn = hertz_to_nm
        find_matching_spectral_lines_fn = find_matching_spectral_lines

    peaks_ang = [hertz_to_nm_fn(x) * 10 for x in data]
    res = find_matching_spectral_lines_fn(df, peaks_ang, tolerance=tolerance)
    elements_count = res["element"].value_counts()
    elements_final = elements_count.index.tolist()[:3]  # take the three most common elements

    elements_mapped = []
    spectrum_regions_mapped = []
    types_mapped = []

    for element in elements_final:
        res_filtered = res[res["element"] == element]

        # Here we are making sure to only get one spectrum region and type per element
        spectrum_region = res_filtered["spectrum_region"].iloc[0]
        type_ = res_filtered["type"].iloc[0]

        elements_mapped.append(element)
        spectrum_regions_mapped.append(spectrum_region)
        types_mapped.append(type_)

    return elements_mapped, spectrum_regions_mapped, types_mapped



from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict

from goofi.connection import Connection
from goofi.data import Data


class MessageType(Enum):
    """
    The type of a message. The message type determines the content of the message.

    - `ADD_OUTPUT_PIPE`: Sent by the manager to a node to add an output pipe to an output slot.
        - `slot_name_out` (str): The name of the output slot.
        - `slot_name_in` (str): The name of the input slot.
        - `node_connection` (Connection): The connection object to the node. If None, connect to itself.
    - `REMOVE_OUTPUT_PIPE`: Sent by the manager to a node to remove an output pipe from an output slot.
        - `slot_name_out` (str): The name of the output slot.
        - `slot_name_in` (str): The name of the input slot.
        - `node_connection` (Connection): The connection object to the node.
    - `DATA`: Sent by one node to another and contains data sent from an output slot to an input slot.
        - `slot_name` (str): The name of the target input slot.
        - `data` (Data): The data object. See the `Data` class for more information.
    - `CLEAR_DATA`: Clear the data field on an input slot.
    - `PING`: Empty probe to check if a process is alive.
    - `PONG`: Response to a ping message.
    - `TERMINATE`: Empty message to indicate that a node should terminate.
    - `PROCESSING_ERROR`: Sent by a node to indicate that an error occurred during processing.
        - `error` (str): The error message.
    - `PARAMETER_UPDATE`: Sent by the manager to a node to update a parameter.
        - `group` (str): The name of the parameter group.
        - `param_name` (str): The name of the parameter.
        - `param_value` (Any): The new parameter value.
    - `SERIALIZE_REQUEST`: An empty message that requests a node to serialize its state and send it back.
    - `SERIALIZE_RESPONSE`: The response to a serialize request message.
        - `_type` (str): The type of the node.
        - `category` (str): The category of the node.
        - `out_conns` (Dict[str, List[Tuple[str, Connection]]]): A dictionary of named output slots containing connections.
        - `params` (Dict[str, Any]): The NodeParams object, represented as a dictionary.
    """

    ADD_OUTPUT_PIPE = 1
    REMOVE_OUTPUT_PIPE = 2
    DATA = 3
    CLEAR_DATA = 4
    PING = 5
    PONG = 6
    TERMINATE = 7
    PROCESSING_ERROR = 8
    PARAMETER_UPDATE = 9
    SERIALIZE_REQUEST = 10
    SERIALIZE_RESPONSE = 11


@dataclass
class Message:
    """
    A message is used to send information between nodes. Each message has a type and some content. The
    content field is a dict that contains the message content. The content field must contain the correct
    fields for the message type.

    ### Parameters
    `type` : MessageType
        The type of the message.
    `content` : Dict[str, Any]
        The content of the message with required fields for the message type.
    """

    type: MessageType
    content: Dict[str, Any]

    def require_fields(self, **fields: Dict[str, Any]) -> None:
        """
        Check that the message content contains the required fields.

        ### Parameters
        `fields` : Dict[str, Any]
            A dictionary of required fields and their types.
        """
        for field, field_type in fields.items():
            if field not in self.content:
                raise ValueError(f"Message content must contain field {field}.")
            if not isinstance(self.content[field], field_type):
                raise ValueError(
                    f"Message content field {field} must be of type {field_type} but got {type(self.content[field])}."
                )

    def __post_init__(self):
        """
        Check that the message content is valid. The content field must be a dict, and it must contain the correct
        fields for the message type.
        """
        # general checks
        if self.type is None:
            raise ValueError("Expected message type, got None.")
        if not isinstance(self.content, dict):
            raise ValueError(f"Expected dict, got {type(self.content)}.")

        # check the content for the specific message type
        if self.type == MessageType.ADD_OUTPUT_PIPE:
            self.require_fields(slot_name_out=str, slot_name_in=str, node_connection=(Connection, type(None)))
        elif self.type == MessageType.REMOVE_OUTPUT_PIPE:
            self.require_fields(slot_name_out=str, slot_name_in=str, node_connection=Connection)
        elif self.type == MessageType.DATA:
            self.require_fields(slot_name=str, data=Data)
        elif self.type == MessageType.CLEAR_DATA:
            self.require_fields(slot_name=str)
        elif self.type == MessageType.PROCESSING_ERROR:
            self.require_fields(error=(str, type(None)))
        elif self.type == MessageType.PARAMETER_UPDATE:
            self.require_fields(group=str, param_name=str, param_value=object)
        elif self.type == MessageType.SERIALIZE_RESPONSE:
            self.require_fields(_type=str, category=str, out_conns=dict, params=dict)



from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import StringParam


class CardiacRespiration(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {"cardiac": DataType.ARRAY}

    def config_params():
        return {
            "cardiac": {
                "input_type": StringParam("ppg", options=["ppg", "ecg"]),
            }
        }

    def setup(self):
        import neurokit2 as nk

        self.neurokit = nk

    def process(self, data: Data):
        if data is None or data.data is None:
            return None

        if data.data.ndim > 1:
            raise ValueError("Data must be 1D")

        if self.params["cardiac"]["input_type"].value == "ppg":
            raise NotImplementedError("PPG not implemented yet")
            # signal, info = self.neurokit.ppg_process(data.data, sampling_rate=data.meta["sfreq"])
            # hrv_df = self.neurokit.hrv(info, sampling_rate=data.meta["sfreq"])
        elif self.params["cardiac"]["input_type"].value == "ecg":
            # extract peaks
            rpeaks, info = self.neurokit.ecg_peaks(data.data, sampling_rate=data.meta["sfreq"])
            # compute rate
            ecg_rate = self.neurokit.ecg_rate(rpeaks, sampling_rate=data.meta["sfreq"], desired_length=len(data.data))
            edr = self.neurokit.ecg_rsp(ecg_rate, sampling_rate=data.meta["sfreq"])

        return {"cardiac": (edr, data.meta)}



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import BoolParam, FloatParam, IntParam, StringParam


class Connectivity(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {
            "matrix": DataType.ARRAY,
        }

    def config_params():
        return {
            "classical": {
                "method": StringParam(
                    "wPLI",
                    options=["coherence", "imag_coherence", "wPLI", "PLI", "PLV", "covariance", "pearson", "mutual_info"],
                ),
            },
            "biotuner": {
                "method": StringParam(
                    "None", options=["None", "harmsim", "euler", "subharm_tension", "RRCi", "wPLI_crossfreq"]
                ),
                "n_peaks": IntParam(5, 1, 10, doc="Number of peaks to extract"),
                "f_min": FloatParam(2.0, 0.1, 50.0, doc="Minimum frequency"),
                "f_max": FloatParam(30.0, 1.0, 100.0, doc="Maximum frequency"),
                "precision": FloatParam(0.1, 0.01, 10.0, doc="Precision of the peak extraction in Hz"),
                "peaks_function": StringParam(
                    "EMD", options=["EMD", "fixed", "harmonic_recurrence", "EIMC"], doc="Peak extraction function"
                ),
            },
            "Adjacency": {
                "Binarize": BoolParam(False, doc="Binarize the connectivity matrix"),
                "threshold": FloatParam(0.5, 0.0, 1.0, doc="Threshold for binarization"),
            },
        }

    def process(self, data: Data):
        if data is None:
            return None

        data.data = np.squeeze(data.data)
        if data.data.ndim != 2:
            raise ValueError("Data must be 2D")

        if self.params["biotuner"]["method"].value != "None":
            matrix = compute_conn_matrix_single(
                data.data,
                data.meta["sfreq"],
                peaks_function=self.params["biotuner"]["peaks_function"].value,
                min_freq=self.params["biotuner"]["f_min"].value,
                max_freq=self.params["biotuner"]["f_max"].value,
                precision=self.params["biotuner"]["precision"].value,
                n_peaks=self.params["biotuner"]["n_peaks"].value,
                metric=self.params["biotuner"]["method"].value,
            )

        if self.params["biotuner"]["method"].value == "None":
            method = self.params["classical"]["method"].value
            matrix = compute_classical_connectivity(data.data, method)

        binarize = self.params["Adjacency"]["Binarize"].value
        threshold = self.params["Adjacency"]["threshold"].value

        if binarize:
            matrix[matrix < threshold] = 0
            matrix[matrix >= threshold] = 1
        return {"matrix": (matrix, data.meta)}


connectivity_fn = None


def compute_conn_matrix_single(
    data: np.ndarray,
    sfreq: float,
    peaks_function: str = "EMD",
    min_freq: float = 2.0,
    max_freq: float = 45.0,
    precision=0.1,
    n_peaks: int = 5,
    metric: str = "harmsim",
):
    # import the connectivity function here to avoid loading it on startup
    global connectivity_fn
    if connectivity_fn is None:
        from biotuner.harmonic_connectivity import harmonic_connectivity

        connectivity_fn = harmonic_connectivity

    # compute connectivity matrix
    bt_conn = connectivity_fn(
        sf=sfreq,
        data=data,
        peaks_function=peaks_function,
        precision=precision,
        min_freq=min_freq,
        max_freq=max_freq,
        n_peaks=n_peaks,
    )
    bt_conn.compute_harm_connectivity(metric=metric, save=False, graph=False)
    return bt_conn.conn_matrix


hilbert_fn, coherence_fn, pearsonr_fn, mutual_info_regression_fn = None, None, None, None


def compute_classical_connectivity(data, method):
    # import the connectivity function here to avoid loading it on startup
    global hilbert_fn, coherence_fn, pearsonr_fn, mutual_info_regression_fn
    if hilbert_fn is None:
        from scipy.signal import coherence, hilbert
        from scipy.stats import pearsonr
        from sklearn.feature_selection import mutual_info_regression

        hilbert_fn = hilbert
        coherence_fn = coherence
        pearsonr_fn = pearsonr
        mutual_info_regression_fn = mutual_info_regression

    n_channels, n_samples = data.shape
    matrix = np.zeros((n_channels, n_channels))

    if method == "covariance":
        matrix = np.cov(data)
        return matrix
    # TODO : optimize this shit
    for i in range(n_channels):
        for j in range(i, n_channels):  # Only compute upper diagonal
            if i == j:
                matrix[i, j] = 1  # diagonal elements are 1, for coherence/PLV or 0 for others, adjust as needed
                continue

            if method == "wPLI":
                sig1 = hilbert_fn(data[i, :])
                sig2 = hilbert_fn(data[j, :])
                imag_csd = np.imag(np.exp(1j * (np.angle(sig1) - np.angle(sig2))))
                matrix[i, j] = matrix[j, i] = np.abs(np.mean(imag_csd)) / np.mean(np.abs(imag_csd))

            elif method == "coherence":
                f, Cxy = coherence_fn(data[i, :], data[j, :])
                matrix[i, j] = matrix[j, i] = np.mean(Cxy)

            elif method == "PLI":
                sig1 = hilbert_fn(data[i, :])
                sig2 = hilbert_fn(data[j, :])
                matrix[i, j] = matrix[j, i] = np.mean(np.sign(np.angle(sig1) - np.angle(sig2)))

            elif method == "imag_coherence":
                sig1 = hilbert_fn(data[i, :])
                sig2 = hilbert_fn(data[j, :])
                matrix[i, j] = matrix[j, i] = np.mean(np.imag(np.conj(sig1) * sig2)) / (
                    np.sqrt(np.mean(np.imag(sig1) ** 2)) * np.sqrt(np.mean(np.imag(sig2) ** 2))
                )

            elif method == "PLV":
                sig1 = hilbert_fn(data[i, :])
                sig2 = hilbert_fn(data[j, :])
                matrix[i, j] = matrix[j, i] = np.abs(np.mean(np.exp(1j * (np.angle(sig1) - np.angle(sig2)))))

            elif method == "pearson":
                corr, _ = pearsonr_fn(data[i, :], data[j, :])
                matrix[i, j] = matrix[j, i] = corr

            elif method == "mutual_info":
                mutual_info = mutual_info_regression_fn(data[i, :].reshape(-1, 1), data[j, :])[0]
                matrix[i, j] = matrix[j, i] = mutual_info

    return matrix



from os import path

import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node


class PoseEstimation(Node):
    def config_input_slots():
        return {"image": DataType.ARRAY}

    def config_output_slots():
        return {"pose": DataType.ARRAY}

    def setup(self):
        import mediapipe as mp
        import requests
        from mediapipe.tasks import python
        from mediapipe.tasks.python import vision

        self.mp = mp

        model_path = path.join(self.assets_path, "hand_landmarker.task")
        if not path.exists(model_path):
            url = (
                "https://storage.googleapis.com/mediapipe-models/hand_landmarker/hand_landmarker/float16/1/hand_landmarker.task"
            )
            response = requests.get(url)
            # ensure the request was successful
            response.raise_for_status()
            # write the content to a file
            with open(model_path, "wb") as file:
                file.write(response.content)

        base_options = python.BaseOptions(model_asset_path=model_path)
        options = vision.HandLandmarkerOptions(base_options=base_options, num_hands=1)
        self.detector = vision.HandLandmarker.create_from_options(options)

    def process(self, image: Data):
        mp_image = self.mp.Image(image_format=self.mp.ImageFormat.SRGB, data=(image.data * 255).astype(np.uint8))
        detection_result = self.detector.detect(mp_image)

        if len(detection_result.hand_landmarks) == 0:
            return None

        landmarks = np.array([[lm.x, lm.y, lm.z] for lm in detection_result.hand_landmarks[0]]).T
        meta = {
            "handdedness": detection_result.handedness[0][0].display_name,
            "channels": {"dim0": ["x", "y", "z"], "dim1": LANDMARK_NAMES},
        }
        return {"pose": (landmarks, meta)}


LANDMARK_NAMES = [
    "WRIST",
    "THUMB_CMC",
    "THUMB_MCP",
    "THUMB_IP",
    "THUMB_TIP",
    "INDEX_FINGER_MCP",
    "INDEX_FINGER_PIP",
    "INDEX_FINGER_DIP",
    "INDEX_FINGER_TIP",
    "MIDDLE_FINGER_MCP",
    "MIDDLE_FINGER_PIP",
    "MIDDLE_FINGER_DIP",
    "MIDDLE_FINGER_TIP",
    "RING_FINGER_MCP",
    "RING_FINGER_PIP",
    "RING_FINGER_DIP",
    "RING_FINGER_TIP",
    "PINKY_MCP",
    "PINKY_PIP",
    "PINKY_DIP",
    "PINKY_TIP",
]



from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import BoolParam, StringParam


class Audiocraft(Node):
    def config_params():
        return {
            "audiocraft": {
                "model": StringParam(
                    "facebook/audiogen-medium",
                    options=[
                        "facebook/audiogen-medium",
                        "facebook/musicgen-small",
                        "facebook/musicgen-medium",
                        "facebook/musicgen-large",
                    ],
                ),
                "device": StringParam("cuda", options=["cuda", "cpu"]),
            },
            "setup": {"install": BoolParam(False, trigger=True)},
        }

    def config_input_slots():
        return {"prompt": DataType.STRING}

    def config_output_slots():
        return {"wav": DataType.ARRAY}

    def setup(self):
        try:
            from audiocraft.models import AudioGen, MusicGen
        except ImportError:
            raise ImportError(
                "Install audiocraft using the 'install' button or according to the instructions here: "
                "https://github.com/facebookresearch/audiocraft/tree/main?tab=readme-ov-file#installation"
            )

        try:
            import torch
        except ImportError:
            raise ImportError(
                "Please first install torch according to these instructions: https://pytorch.org/get-started/locally/"
            )

        torch.set_grad_enabled(False)

        self.current_model = self.params.audiocraft.model.value

        if "audiogen" in self.params.audiocraft.model.value:
            self.model = AudioGen.get_pretrained(self.params.audiocraft.model.value, device=self.params.audiocraft.device.value)
        elif "musicgen" in self.params.audiocraft.model.value:
            self.model = MusicGen.get_pretrained(self.params.audiocraft.model.value, device=self.params.audiocraft.device.value)
        else:
            raise ValueError(f"Model {self.params.audiocraft.model.value} not found")

    def process(self, prompt: Data):
        wav = self.model.generate([prompt.data])[0].cpu().numpy()
        prompt.meta["sampling_rate"] = self.model.sample_rate
        return {"wav": (wav, prompt.meta)}

    def audiocraft_model_changed(self, model):
        if model != self.current_model:
            self.setup()

    def setup_install_changed(self, install):
        if not install:
            return

        import subprocess
        import tempfile

        import requests

        req = requests.get("https://raw.githubusercontent.com/facebookresearch/audiocraft/main/requirements.txt").text

        # remove version from torch to avoid messing up the current installation
        req = req.replace("torch==2.1.0\n", "")
        try:
            import torch  # noqa
        except ImportError:
            raise ImportError(
                "Please first install torch according to these instructions: https://pytorch.org/get-started/locally/"
            )

        # remove xformers to install independently to avoid not finding a compatible version
        req = req.replace("xformers<0.0.23\n", "")

        # write requirements to a temp file and install them
        with tempfile.NamedTemporaryFile(delete=False) as f:
            f.write(req.encode())
            req_file = f.name
        subprocess.run(["pip", "install", "-r", req_file])

        # install xformers and torchmetrics independently
        subprocess.run(["pip", "install", "xformers", "torchmetrics"])

        # update torchvision and torchaudio if new torch was installed
        subprocess.run(["pip", "install", "-U", "torchvision", "torchaudio"])

        # install audiocraft without dependencies
        subprocess.run(["pip", "install", "audiocraft", "--no-deps"])



import pickle
from os.path import join

import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import BoolParam, FloatParam, IntParam


class AudioTagging(Node):
    def config_input_slots():
        return {"audioIn": DataType.ARRAY}

    def config_output_slots():
        return {"tags": DataType.STRING, "probabilities": DataType.ARRAY, "embedding": DataType.ARRAY}

    def config_params():
        return {
            "selection": {
                "threshold": FloatParam(0.5, 0.0, 1.0),
                "number_of_tags": IntParam(3, 1, 527),
                "all_categories": BoolParam(True),
                "human_sounds": BoolParam(False),
                "animal": BoolParam(False),
                "music": BoolParam(False),
                "sounds_of_things": BoolParam(False),
                "natural_sounds": BoolParam(False),
                "source_ambiguous_sounds": BoolParam(False),
                "channel_environment_and_background": BoolParam(False),
            },
        }

    def setup(self):
        from panns_inference import AudioTagging, labels

        self.at = AudioTagging(checkpoint_path=None, device="cpu")
        self.labels = labels
        audio_tags_path = join(self.assets_path, "audio_tags_structure.pkl")
        with open(audio_tags_path, "rb") as f:
            self.audio_tags = pickle.load(f)

    def process(self, audioIn: Data):
        if audioIn.data is None:
            return None

        if audioIn.meta["sfreq"] != 32000:
            raise ValueError("Sampling frequency must be 32k")

        threshold = self.params["selection"]["threshold"].value
        n_tags = self.params["selection"]["number_of_tags"].value
        audioIn.data = audioIn.data.reshape(1, -1)

        (clipwise_output, embedding) = self.at.inference(audioIn.data)

        best_labels = np.argsort(clipwise_output[0])[::-1]
        tags = np.array(self.labels)[best_labels]
        probabilities = clipwise_output[0][best_labels]
        embedding = embedding.squeeze()

        if self.params["selection"]["all_categories"].value:
            # If all_categories is set to True, include all tags
            active_categories = [tag for sublist in self.audio_tags.values() for tag in sublist]
        else:
            active_categories = []
            if self.params["selection"]["human_sounds"].value:
                active_categories.extend(self.audio_tags["Human sounds"])
            if self.params["selection"]["animal"].value:
                active_categories.extend(self.audio_tags["Animal"])
            if self.params["selection"]["music"].value:
                active_categories.extend(self.audio_tags["Music"])
            if self.params["selection"]["sounds_of_things"].value:
                active_categories.extend(self.audio_tags["Sounds of things"])
            if self.params["selection"]["natural_sounds"].value:
                active_categories.extend(self.audio_tags["Natural sounds"])
            if self.params["selection"]["source_ambiguous_sounds"].value:
                active_categories.extend(self.audio_tags["Source-ambiguous sounds"])
            if self.params["selection"]["channel_environment_and_background"].value:
                active_categories.extend(self.audio_tags["Channel, environment and background"])

        # Filter based on active categories
        indices = [i for i, x in enumerate(tags) if x in active_categories and probabilities[i] > threshold]
        formatted_tags = "\n".join(tags[i] for i in indices[: n_tags - 1])
        formatted_probs = np.array([probabilities[i] for i in indices[: n_tags - 1]])
        
        return {
            "tags": (formatted_tags, audioIn.meta),
            "probabilities": (formatted_probs, audioIn.meta),
            "embedding": (embedding, audioIn.meta),
        }



import pytest

from goofi.gui.window import DTYPE_SHAPE_MAP

from .utils import list_data_types


@pytest.mark.parametrize("dtype", list_data_types())
def test_dtype_shape_map(dtype):
    assert dtype in DTYPE_SHAPE_MAP, f"{dtype} not in DTYPE_SHAPE_MAP."



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import FloatParam, IntParam, BoolParam


class Math(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {"out": DataType.ARRAY}

    def config_params():
        return {
            "math": {
                "pre_add": FloatParam(0.0, -10.0, 10.0),
                "multiply": FloatParam(1.0, -10.0, 10.0),
                "post_add": FloatParam(0.0, -10.0, 10.0),
                "round": IntParam(-1, -1, 10),
                "sqrt": BoolParam(False),
                "squared": BoolParam(False),
            },
            "map": {
                "input_min": FloatParam(0.0, -10.0, 10.0),
                "input_max": FloatParam(1.0, -10.0, 10.0),
                "output_min": FloatParam(0.0, -10.0, 10.0),
                "output_max": FloatParam(1.0, -10.0, 10.0),
            },
        }

    def process(self, data: Data):
        if data is None or data.data is None:
            return None

        signal = data.data

        # apply math operations
        signal += self.params["math"]["pre_add"].value
        signal *= self.params["math"]["multiply"].value
        signal += self.params["math"]["post_add"].value

        # apply rounding
        decimals = self.params["math"]["round"].value
        if decimals >= 0:
            signal = np.around(signal, decimals)

        # rescale signal from input range to output range
        signal = self.rescale(
            signal,
            self.params["map"]["input_min"].value,
            self.params["map"]["input_max"].value,
            self.params["map"]["output_min"].value,
            self.params["map"]["output_max"].value,
        )
        if self.params["math"]["sqrt"].value:
            signal = np.sqrt(signal)
        if self.params["math"]["squared"].value:
            signal = np.square(signal)
        return {"out": (signal, data.meta)}

    @staticmethod
    def rescale(signal, input_min, input_max, output_min, output_max):
        # rescale signal from input range to output range
        return ((signal - input_min) / (input_max - input_min)) * (output_max - output_min) + output_min



import base64
import io

import numpy as np
from PIL import Image

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import FloatParam, IntParam, StringParam


class Img2Txt(Node):
    @staticmethod
    def config_input_slots():
        return {"image": DataType.ARRAY}

    @staticmethod
    def config_output_slots():
        return {"generated_text": DataType.STRING}

    @staticmethod
    def config_params():
        return {
            "img_to_text": {
                "model": StringParam(
                    "gpt-4o-mini",
                    options=["ollama:llama3.2-vision", "meta-llama/Llama-3.2-11B-Vision-Instruct", "gpt-4o-mini", "ollama"],
                    doc="Model ID or name for image captioning (Huggingface Llama, OpenAI, or Ollama)",
                ),
                "max_new_tokens": IntParam(30, 10, 1024, doc="Maximum number of tokens to generate"),
                "temperature": FloatParam(0.7, 0.1, 2.0, doc="Sampling temperature"),
                "prompt": StringParam("What is in this image?", doc="Prompt for image captioning"),
                "openai_key": StringParam("openai.key", doc="Path to OpenAI API key file (required for OpenAI models)"),
            }
        }

    def setup(self):
        import requests

        self.requests = requests

        self.processor = None
        self.model_instance = None
        self.openai = None
        self.ollama = None
        self.model_id = self.params.img_to_text.model.value

        if "/" in self.model_id.lower():
            self.setup_huggingface_llama()
        elif "gpt" in self.model_id.lower():
            self.setup_openai_gpt()
        elif "ollama" in self.model_id.lower():
            self.setup_ollama()
        else:
            raise ValueError(f"Invalid model selection: {self.model_id}")

    def setup_huggingface_llama(self):
        try:
            import torch
            from transformers import AutoProcessor, MllamaForConditionalGeneration

            self.model_instance = MllamaForConditionalGeneration.from_pretrained(
                self.model_id,
                torch_dtype=torch.bfloat16,
                device_map="auto",
            )
            self.processor = AutoProcessor.from_pretrained(self.model_id)
        except Exception as e:
            print(f"Error loading Llama model: {e}")
            raise

    def setup_openai_gpt(self):
        try:
            import openai

            self.openai = openai
            key = self.params["img_to_text"]["openai_key"].value
            with open(key, "r") as f:
                self.openai.api_key = f.read().strip()
        except ImportError:
            print("Error: 'openai' library not found. Please install it using 'pip install openai'.")
            raise
        except FileNotFoundError:
            print(f"Error: OpenAI API key file not found at path: {key}")
            raise
        except Exception as e:
            print(f"Error initializing OpenAI: {e}")
            raise

    def setup_ollama(self):
        try:
            import ollama

            self.ollama = ollama
        except ImportError:
            print("Error: 'ollama' library not found. Please install it using 'pip install ollama'.")
            raise
        except ModuleNotFoundError:
            print("Error: 'ollama' library not found. Please install it using 'pip install ollama'.")
            raise

    def encode_image(self, image_array):
        if image_array.dtype != "uint8":
            image_array = (255 * (image_array - image_array.min()) / (image_array.max() - image_array.min())).astype("uint8")
        image = Image.fromarray(image_array)
        buffer = io.BytesIO()
        image.save(buffer, format="JPEG")
        return base64.b64encode(buffer.getvalue()).decode("utf-8")

    def process(self, image: Data):
        if image.data is None:
            return None

        image_array = image.data

        # Handle various image array shapes and types
        if image_array.ndim == 4:
            if image_array.shape[0] == 1:  # Remove batch dimension if present
                image_array = image_array.squeeze(0)
            else:
                raise ValueError(f"Unexpected image array shape: {image_array.shape}")

        if image_array.ndim == 3:
            if image_array.shape[2] == 3:  # RGB image
                pass
            elif image_array.shape[2] == 1:  # Grayscale image
                image_array = np.repeat(image_array, 3, axis=2)
            else:
                raise ValueError(f"Unexpected number of channels: {image_array.shape[2]}")
        elif image_array.ndim == 2:  # Grayscale image without channel dimension
            image_array = np.stack([image_array] * 3, axis=-1)
        else:
            raise ValueError(f"Unexpected image array dimensions: {image_array.ndim}")

        # Ensure the array is in uint8 format
        if image_array.dtype != np.uint8:
            if np.issubdtype(image_array.dtype, np.floating):
                image_array = (image_array * 255).astype(np.uint8)
            else:
                image_array = image_array.astype(np.uint8)

        if "/" in self.model_id.lower():
            return self.process_huggingface_llama(image_array)
        elif "gpt" in self.model_id.lower():
            return self.process_openai_gpt(image_array)
        elif "ollama" in self.model_id.lower():
            return self.process_ollama(image_array)

    def process_huggingface_llama(self, image_array):
        if self.processor is None or self.model_instance is None:
            return {"generated_text": ("Error: Huggingface Llama model not initialized.", {})}

        # Process using Huggingface Llama
        base64_image = self.encode_image(image_array)
        messages = [
            [
                {
                    "role": "user",
                    "content": [{"type": "image"}, {"type": "text", "text": self.params.img_to_text.prompt.value}],
                }
            ],
        ]
        text = self.processor.apply_chat_template(messages, add_generation_prompt=True)

        # Wrap the image in a list
        inputs = self.processor(text=text, images=[base64_image], return_tensors="pt")

        # Move inputs to the correct device
        inputs = {k: v.to(self.model_instance.device) for k, v in inputs.items()}

        output = self.model_instance.generate(
            **inputs,
            max_new_tokens=self.params.img_to_text.max_new_tokens.value,
            temperature=self.params.img_to_text.temperature.value,
        )
        generated_text = self.processor.decode(output[0], skip_special_tokens=True)
        return {"generated_text": (generated_text, {})}

    def process_openai_gpt(self, image_array):
        # Process using OpenAI GPT
        base64_image = self.encode_image(image_array)
        headers = {"Content-Type": "application/json", "Authorization": f"Bearer {self.openai.api_key}"}
        payload = {
            "model": self.model_id,
            "messages": [
                {
                    "role": "user",
                    "content": [
                        {"type": "text", "text": self.params.img_to_text.prompt.value},
                        {"type": "image_url", "image_url": {"url": f"data:image/jpeg;base64,{base64_image}"}},
                    ],
                }
            ],
            "max_tokens": self.params.img_to_text.max_new_tokens.value,
            "temperature": self.params.img_to_text.temperature.value,
        }

        try:
            response = self.requests.post("https://api.openai.com/v1/chat/completions", headers=headers, json=payload)
            response.raise_for_status()
            generated_text = response.json().get("choices", [{}])[0].get("message", {}).get("content", "")
        except self.requests.exceptions.RequestException as e:
            print(f"Error during OpenAI captioning request: {e}")
            return {"generated_text": ("Error generating caption.", {})}

        return {"generated_text": (generated_text, {})}

    def process_ollama(self, image_array):
        # Process using Ollama
        base64_image = self.encode_image(image_array)
        messages = [{"role": "user", "content": self.params.img_to_text.prompt.value, "images": [base64_image]}]
        response = self.ollama.chat(
            model=self.model_id.replace("ollama:", ""),
            messages=messages,
            options={
                "num_predict": self.params.img_to_text.max_new_tokens.value,
                "temperature": self.params.img_to_text.temperature.value,
            },
        )
        generated_text = response["message"]["content"]

        return {"generated_text": (generated_text, {})}

    def img_to_text_model_changed(self, model_id):
        # reinitialize the model when the model_id changes
        if self.model_id != model_id:
            self.setup()



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import BoolParam, FloatParam, IntParam


class Biorhythms(Node):
    def config_input_slots():
        return {"tuning": DataType.ARRAY}

    def config_output_slots():
        return {
            "pulses": DataType.ARRAY,
            "steps": DataType.ARRAY,
            "offsets": DataType.ARRAY,
        }

    def config_params():
        return {
            "Euclidean": {
                "n_steps_down": IntParam(2, 0, 10, doc="Number of time to divide the ratio"),
                "limit_denom": IntParam(8, 1, 100, doc="Maximum denominator of the ratio"),
                "limit_cons": FloatParam(0.5, 0.0, 1.0, doc="Minimum consonance of the ratio"),
                "limit_denom_final": IntParam(8, 1, 100, doc="Maximum denominator of the final ratio"),
                "optimize_offset": BoolParam(False),
            }
        }

    def setup(self):
        from biotuner.rhythm_construction import consonant_euclid, find_optimal_offsets

        self.consonant_euclid = consonant_euclid
        self.find_optimal_offsets = find_optimal_offsets

    def process(self, tuning: Data):
        if tuning is None:
            return None

        tuning.data = np.squeeze(tuning.data)
        if tuning.data.ndim > 1:
            raise ValueError("Data must be 1D")

        n_steps_down = self.params["Euclidean"]["n_steps_down"].value
        limit_denom = self.params["Euclidean"]["limit_denom"].value
        limit_cons = self.params["Euclidean"]["limit_cons"].value
        limit_denom_final = self.params["Euclidean"]["limit_denom_final"].value
        optimize_offset = self.params["Euclidean"]["optimize_offset"].value
        # Derive consonant euclidian rhythms from the harmonic tuning
        euclid_final, cons = self.consonant_euclid(
            list(tuning.data),
            n_steps_down=n_steps_down,
            limit_denom=limit_denom,
            limit_cons=limit_cons,
            limit_denom_final=limit_denom_final,
        )
        # Calculate the pulses and steps
        pulses = []
        steps = []
        for i in range(len(euclid_final)):
            pulses.append(euclid_final[i].count(1))
            steps.append(len(euclid_final[i]))

        if optimize_offset:
            # find the optimal offsets
            offsets = self.find_optimal_offsets(list(zip(pulses, steps)))
        else:
            offsets = [0] * len(pulses)

        return {
            "pulses": (np.array(pulses), tuning.meta),
            "steps": (np.array(steps), tuning.meta),
            "offsets": (np.array(offsets), tuning.meta),
        }



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import FloatParam


class FrequencyShift(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {"out": DataType.ARRAY}

    def config_params():
        return {
            "shift": {
                "frequency_shift": FloatParam(1.0, -1000.0, 1000.0, doc="Frequency shift in Hz"),
            }
        }

    def setup(self):
        from scipy.signal import istft, stft

        self.istft = istft
        self.stft = stft

    def process(self, data: Data):
        if data is None or data.data is None:
            return None

        signal = np.array(data.data).flatten()  # Assuming the input is 1D for simplicity
        sfreq = data.meta["sfreq"]

        # Perform STFT
        f, t, Zxx = self.stft(signal, fs=sfreq, nperseg=1024)

        # Frequency shifting
        frequency_shift = self.params["shift"]["frequency_shift"].value
        delta_bin = -int(frequency_shift * len(f) / sfreq)  # Negate delta_bin computation
        target_bins = np.arange(len(f)) + delta_bin

        # Handling bins that go out of bounds after the shift
        valid_bins = (target_bins >= 0) & (target_bins < len(f))
        Zxx_shifted = np.zeros_like(Zxx)

        # Phase correction and bin shifting
        phase_shift = (2 * np.pi * delta_bin * np.arange(Zxx.shape[1]) / (Zxx.shape[1])) % (2 * np.pi)
        Zxx_shifted[valid_bins] = Zxx[target_bins[valid_bins]] * np.exp(1j * phase_shift)

        # Inverse STFT to retrieve shifted signal
        _, shifted_signal = self.istft(Zxx_shifted, fs=sfreq)

        # Depending on the exact length and shifts, you might want to match the output size to the input size
        shifted_signal = shifted_signal[: len(signal)]

        return {"out": (shifted_signal, data.meta)}



import pytest
import yaml

from goofi import params
from goofi.params import DEFAULT_PARAMS, BoolParam, FloatParam, NodeParams

from .utils import list_param_types


def full_node_params(n_groups: int = 2) -> NodeParams:
    """
    Returns a NodeParams object with all parameter types.

    ### Parameters
    `n_groups` : int
        The number of parameter groups to create.

    ### Returns
    NodeParams
        A NodeParams object with all parameter types.
    """
    params = {}
    for i in range(n_groups):
        params[f"group{i}"] = {}
        for param_type in list_param_types():
            params[f"group{i}"][f"param{param_type.__name__}"] = param_type.default()
    return NodeParams(params)


def test_default_members():
    assert "common" in DEFAULT_PARAMS, "Missing default group 'common'."
    # autotrigger
    assert "autotrigger" in DEFAULT_PARAMS["common"], "Missing default param 'autotrigger' in group 'common'."
    assert isinstance(
        DEFAULT_PARAMS["common"]["autotrigger"], params.BoolParam
    ), "Wrong type for default param 'autotrigger' in group 'common'"
    # max_frequency
    assert "max_frequency" in DEFAULT_PARAMS["common"], "Missing default param 'max_frequency' in group 'common'."
    assert isinstance(
        DEFAULT_PARAMS["common"]["max_frequency"], params.FloatParam
    ), "Wrong type for default param 'max_frequency' in group 'common'."


@pytest.mark.parametrize("param_type", list_param_types())
def test_param_types(param_type):
    p = param_type()
    assert p.value == param_type.default(), f"Wrong default value for param type {param_type.__name__}."

    for type2 in list_param_types():
        if type2 is param_type:
            continue

        if isinstance(type2.default(), type(param_type.default())):
            # some types may be instances of other types, e.g. bool is an instance of int
            continue

        if isinstance(param_type.default(), float) and isinstance(type2.default(), int):
            # we accept int as default value for float
            continue

        with pytest.raises(TypeError):
            param_type(type2.default())


def test_node_params():
    params = NodeParams(DEFAULT_PARAMS)

    try:
        params.common
    except AttributeError:
        pytest.fail("NodeParams should support attribute access (e.g. params.common).")
    try:
        params.common.autotrigger
    except AttributeError:
        pytest.fail("Parameter groups should support attribute access (e.g. params.common.autotrigger).")

    with pytest.raises(AttributeError):
        params.nonexistent

    for name, group in DEFAULT_PARAMS.items():
        assert getattr(params, name)._asdict() == group, f"Wrong parameter group {name} in NodeParams."


def test_node_params_repr():
    repr(NodeParams(DEFAULT_PARAMS))


@pytest.mark.parametrize("param_type", list_param_types())
def test_type_param_map(param_type):
    assert any(
        param_type == t for t in params.TYPE_PARAM_MAP.values()
    ), f"Param type '{param_type.__name__}' is not in TYPE_PARAM_MAP."


@pytest.mark.parametrize("type_cls", params.TYPE_PARAM_MAP.keys())
def test_raw_value(type_cls):
    data = {"test": {"param": type_cls()}}
    NodeParams(data)


def test_node_params_iter():
    params = DEFAULT_PARAMS.copy()
    params["test"] = {"param": True}

    p = NodeParams(params)
    assert len(p) == 2, "NodeParams should have 2 groups."

    it = iter(p)
    assert next(it) == "common", "First group should be 'common'."
    assert next(it) == "test", "Second group should be 'test'."
    with pytest.raises(StopIteration):
        next(it)

    expected = ["common", "test"]
    groups = list(p)
    assert groups == expected, "NodeParams should have groups 'common' and 'test'."

    for i, group in enumerate(p):
        assert group == expected[i], f"NodeParams group {i} should be '{expected[i]}'."


def test_node_params_contains():
    params = DEFAULT_PARAMS.copy()
    params["test"] = {"param": True}

    p = NodeParams(params)
    assert "common" in p, "NodeParams should contain group 'common'."
    assert "test" in p, "NodeParams should contain group 'test'."
    assert "nonexistent" not in p, "NodeParams should not contain group 'nonexistent'."


def test_node_params_getitem():
    p = NodeParams(DEFAULT_PARAMS)
    assert p[0] == "common", "NodeParams group 0 should be 'common'."
    assert p["common"] == p.common, "NodeParams group 'common' should be accessible by name."


def test_param_group_getitem():
    p = NodeParams(DEFAULT_PARAMS)
    assert p.common["autotrigger"].value is False, "NodeParams group 'common' should have the correct value for 'autotrigger'."


def test_param_group_keys():
    p = NodeParams(DEFAULT_PARAMS)
    assert p.common.keys() == DEFAULT_PARAMS["common"].keys(), "NodeParams group 'common' should have the correct keys."


def test_param_group_values():
    p = NodeParams(DEFAULT_PARAMS)
    assert isinstance(
        p.common.values(), type(DEFAULT_PARAMS["common"].values())
    ), "NodeParams group 'common' should have a dict_values object."
    assert list(p.common.values()) == list(
        DEFAULT_PARAMS["common"].values()
    ), "NodeParams group 'common' should have the correct values."


def test_param_group_items():
    p = NodeParams(DEFAULT_PARAMS)
    assert p.common.items() == DEFAULT_PARAMS["common"].items(), "NodeParams group 'common' should have the correct items."


def test_serialize():
    p = full_node_params()

    # make sure the serialized data can be converted to YAML
    serialized = p.serialize()
    serialized_str = yaml.dump(serialized)

    # reconstruct from the serialized data
    reconstructed = yaml.load(serialized_str, Loader=yaml.FullLoader)
    p2 = NodeParams(reconstructed)

    # make sure the reconstructed object is equal to the original
    assert p == p2, "NodeParams should be equal after serialization and reconstruction."


def test_update():
    p = full_node_params()
    p.update({"group0": {"paramFloatParam": 5.0}})
    assert p.group0.paramFloatParam.value == 5.0, "NodeParams should update the value of a parameter."

    p.update({"group0": {"paramFloatParam": FloatParam(-3.0)}})
    assert p.group0.paramFloatParam.value == -3.0, "NodeParams should update the value of a parameter."


@pytest.mark.parametrize("param_type", list_param_types())
def test_update_types(param_type):
    p = full_node_params()
    p.update({"group0": {"param" + param_type.__name__: param_type.default()}})


@pytest.mark.parametrize("param_type", list_param_types())
def test_doc(param_type):
    p = param_type()
    assert p.doc is None, f"Param type {param_type.__name__} should have no doc string."

    p = param_type(doc="test")
    assert p.doc == "test", f"Param type {param_type.__name__} should have the correct doc string."

    p = param_type(param_type.default(), doc="test")
    assert p.doc == "test", f"Param type {param_type.__name__} should have the correct doc string."


@pytest.mark.parametrize("trigger", [True, False])
def test_toggle_bool_param(trigger):
    p = BoolParam(True, trigger=trigger)
    assert p.value, "BoolParam should have the correct value."
    assert p.value == (not trigger), "Accessing BoolParam should set the value to False if trigger is True."



from goofi.data import Data, DataType
from goofi.node import Node


class AppendTables(Node):
    def config_input_slots():
        return {"table1": DataType.TABLE, "table2": DataType.TABLE}

    def config_output_slots():
        return {"output_table": DataType.TABLE}

    def config_params():
        return {}

    def process(self, table1: Data, table2: Data):
        if table1 is None and table2 is None:
            return None
        if table2 is None:
            return {"output_table": (table1.data, table1.meta)}
        if table1 is None:
            return {"output_table": (table2.data, table2.meta)}

        # Combine the two tables' data dictionaries
        combined_data = {**table1.data, **table2.data}

        # For simplicity, I'm just combining the meta data dictionaries here,
        # but you might want to handle meta data merging more carefully.
        combined_meta = {**table1.meta, **table2.meta}

        return {"output_table": (combined_data, combined_meta)}



import socket  # For setting broadcast options
from typing import Any, List, Tuple

from oscpy.client import send_bundle, send_message

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import BoolParam, IntParam


class OSCOut(Node):
    def config_input_slots():
        return {"data": DataType.TABLE}

    def config_params():
        return {
            "osc": {
                "address": "localhost",
                "port": IntParam(8000, 0, 65535),
                "prefix": "/goofi",
                "bundle": BoolParam(False, doc="Some software doesn't deal well with OSC bundles"),
                "broadcast": BoolParam(False, doc="Enable broadcasting OSC messages"),
            }
        }

    def setup(self):
        self.sock = None

    def process(self, data: Data):
        if data is None or len(data.data) == 0:
            return

        # determine the address and configure the socket if broadcasting is enabled
        address = self.params.osc.address.value
        port = self.params.osc.port.value
        broadcast = self.params.osc.broadcast.value

        if broadcast:
            address = "255.255.255.255"
            if self.sock is None:
                # create a socket with broadcasting enabled if needed
                self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

        # convert the data to a list of OSC messages
        messages = generate_messages(data, self.params.osc.prefix.value)

        if self.params.osc.bundle.value:
            # send the data as an OSC bundle
            send_bundle(messages, address, port, sock=self.sock)
        else:
            # send the data as individual OSC messages
            for addr, val in messages:
                if self.sock is None:
                    send_message(addr, val, address, port)
                else:
                    send_message(addr, val, address, port, sock=self.sock)

    def osc_broadcast_changed(self, value: bool):
        if self.sock is not None:
            self.sock.close()
            self.sock = None


def generate_messages(data: Data, prefix: str = "") -> List[Tuple[bytes, List[Any]]]:
    messages = []
    for key, val in data.data.items():
        # generate the OSC address
        addr = sanitize_address(prefix + "/" + key)

        if val.dtype == DataType.ARRAY:
            # convert the array to a list
            assert val.data.ndim < 2, "Numerical arrays must at most be one-dimensional."
            val = val.data.tolist()
        elif val.dtype == DataType.STRING:
            # simply use the string
            val = val.data.encode("utf-8")
        elif val.dtype == DataType.TABLE:
            # recursively convert the table to a list of messages
            messages.extend(generate_messages(val, addr))
            continue
        else:
            raise ValueError(f"Unsupported data type {val.dtype} for OSC output.")

        # oscpy expects the message to be a list
        if not isinstance(val, list):
            val = [val]

        # add the message to the list
        messages.append((addr.encode(), val))

    return messages


def sanitize_address(address: str) -> str:
    """
    Sanitize an OSC address. This function removes leading and trailing slashes and replaces multiple slashes with a
    single slash.

    ### Parameters
    `address` : str
        The OSC address to sanitize.

    ### Returns
    str
        The sanitized OSC address.
    """
    return "/" + "/".join(a for a in address.split("/") if len(a) > 0).strip("/")



import threading

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import IntParam, StringParam


class MidiCCout(Node):
    def config_input_slots():
        return {
            "cc1": DataType.ARRAY,
            "cc2": DataType.ARRAY,
            "cc3": DataType.ARRAY,
            "cc4": DataType.ARRAY,
            "cc5": DataType.ARRAY,
        }

    def config_output_slots():
        return {"midi_status": DataType.STRING}

    def config_params():
        try:
            import mido

            available_ports = mido.get_output_names()
        except Exception as e:
            print(f"Error getting MIDI ports: {e}")
            available_ports = []

        return {
            "MIDI": {
                "port_name": StringParam("", options=available_ports, doc="MIDI output port name"),
                "channel": IntParam(1, 1, 15, doc="MIDI channel"),
                "cc1": IntParam(1, 0, 127, doc="MIDI CC number for input 1"),
                "cc2": IntParam(2, 0, 127, doc="MIDI CC number for input 2"),
                "cc3": IntParam(3, 0, 127, doc="MIDI CC number for input 3"),
                "cc4": IntParam(4, 0, 127, doc="MIDI CC number for input 4"),
                "cc5": IntParam(5, 0, 127, doc="MIDI CC number for input 5"),
            }
        }

    def send_cc(self, outport, cc_number, value, channel):
        """Send a MIDI CC message."""
        # convert cc_number to int
        cc_number = int(cc_number)
        value = int(value)
        outport.send(self.mido.Message("control_change", control=cc_number, value=value, channel=channel))

    def setup(self):
        import mido

        self.mido = mido

    def process(self, cc1: Data, cc2: Data, cc3: Data, cc4: Data, cc5: Data):
        port_name = self.params.MIDI["port_name"].value
        channel = self.params.MIDI["channel"].value
        channel = channel - 1  # MIDI channels are 0-indexed
        cc_numbers = [
            self.params.MIDI["cc1"].value,
            self.params.MIDI["cc2"].value,
            self.params.MIDI["cc3"].value,
            self.params.MIDI["cc4"].value,
            self.params.MIDI["cc5"].value,
        ]
        cc_data = [cc1, cc2, cc3, cc4, cc5]

        outport = self.mido.open_output(port_name) if port_name != "goofi" else self.goofi_port

        alert_on = False
        try:
            threads = []
            for i, data in enumerate(cc_data):
                if data is not None and len(data.data) > 0:
                    for value in data.data:
                        if value < 0 or value > 127:
                            print(f"Value outside of MIDI 0-127 range: {value}")
                            alert_on = True
                            error_message = f"Value outside of MIDI 0-127 range: {value}"
                            pass
                        t = threading.Thread(target=self.send_cc, args=(outport, cc_numbers[i], value, channel))
                        t.start()
                        threads.append(t)

            for t in threads:
                t.join()
        finally:
            if port_name != "goofi":
                outport.close()  # Ensure that the MIDI port is closed when done

        if alert_on:
            return {"midi_status": (f"CC messages sent with errors\n{error_message}", cc1.meta)}
        else:
            return {"midi_status": ("CC messages sent successfully", cc1.meta)}



import socket
from typing import Any, Dict, Tuple

import numpy as np

from goofi.data import DataType
from goofi.node import Node
from goofi.params import BoolParam


class LSLClient(Node):
    def config_params():
        return {
            "lsl_stream": {
                "source_name": "goofi",
                "stream_name": "",
                "refresh": BoolParam(False, trigger=True),
            },
            "common": {"autotrigger": True},
        }

    def config_output_slots():
        return {"out": DataType.ARRAY}

    def setup(self):
        """Initialize and start the LSL client."""
        import pylsl

        self.pylsl = pylsl

        if hasattr(self, "client"):
            self.disconnect()
        self.client = None
        self.ch_names = None

        # initialize list of streams
        self.available_streams = None
        self.lsl_stream_refresh_changed(True)

        self.connect()

    def process(self) -> Dict[str, Tuple[np.ndarray, Dict[str, Any]]]:
        """Fetch the next chunk of data from the client."""
        if self.available_streams is None:
            self.lsl_stream_refresh_changed(True)

        if self.client is None:
            if not self.connect():
                return None

        try:
            # fetch data
            samples, _ = self.client.pull_chunk()
        except self.pylsl.LostError:
            self.setup()
            return None

        samples = np.array(samples).T

        if samples.size == 0:
            return None

        try:
            ch_info = self.client.info().desc().child("channels").child("channel")
            ch_type = self.client.info().type().lower()
            ch_names = []
            for k in range(1, self.client.info().channel_count() + 1):
                ch_names.append(ch_info.child_value("label") or "{} {:03d}".format(ch_type.upper(), k))
                ch_info = ch_info.next_sibling()
            self.ch_names = ch_names
        except Exception:
            self.setup()

        meta = {
            "sfreq": self.client.info().nominal_srate(),
            "channels": {"dim0": self.ch_names},
        }
        return {"out": (samples, meta)}

    def connect(self) -> bool:
        """Connect to the LSL stream."""
        self.lsl_stream_refresh_changed(True)
        self.disconnect()

        # find the stream
        source_name = self.params.lsl_stream.source_name.value
        stream_name = self.params.lsl_stream.stream_name.value

        matches = {}
        for info in self.available_streams:
            h, s, n = info.hostname(), info.source_id(), info.name()
            if s == source_name and (len(stream_name) == 0 or n == stream_name):
                if (s, n) in matches and h == socket.gethostname():
                    # prefer local streams
                    matches[(s, n)] = info
                elif (s, n) not in matches:
                    # otherwise, prefer the first match
                    matches[(s, n)] = info

        if len(matches) == 0:
            raise RuntimeError(f'Could not find stream "{stream_name}" from source "{source_name}".')
        elif len(matches) > 1:
            ms = {m[0]: m[1] for m in matches.keys()}
            raise RuntimeError(f'Found multiple streams matching "{stream_name}" from source "{source_name}": {ms}.')

        # connect to the stream
        self.client = self.pylsl.StreamInlet(info=list(matches.values())[0], recover=False)
        return True

    def disconnect(self) -> None:
        """Disconnect from the LSL stream."""
        if self.client is not None:
            self.client.close_stream()
            self.client = None

    def lsl_stream_refresh_changed(self, value: bool) -> None:
        self.available_streams = self.pylsl.resolve_streams()
        print("\nAvailable LSL streams:")
        for info in self.available_streams:
            print(f'  Source: "{info.source_id()}" with stream "{info.name()}" (hostname: {info.hostname()})')
        print()

    def lsl_stream_source_name_changed(self, value: str) -> None:
        assert value != "", "Host name cannot be empty."
        self.setup()

    def lsl_stream_stream_name_changed(self, value: str) -> None:
        assert value != "", "Stream name cannot be empty."
        self.setup()



import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import BoolParam, IntParam


class VAMP(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {
            "comps": DataType.ARRAY,
        }

    def config_params():
        return {
            "vamp": {
                "time_lag": IntParam(10, 1, 300, doc="Time lag tau (in samples)"),
                "n_dims": IntParam(2, 1, 8, doc="Number of dimensions to keep"),
                "collect": BoolParam(True, doc="Collect data and fit the model"),
                "epoch_size": IntParam(256, 2, 1000, doc="Size of one epoch (in samples)"),
                "reset": BoolParam(False, trigger=True, doc="Reset the model"),
            }
        }

    def setup(self):
        self.current_epoch = None
        self.buffer = []
        self.internal_model = self.make_model()
        self.model = None

    def process(self, data: Data):
        if data is None:
            return None

        assert data.data.ndim == 2, "Data must be 2D"

        if self.params.vamp.reset.value:
            self.buffer = []
            self.current_epoch = None
            self.internal_model = self.make_model()
            self.model = None

        if self.params.vamp.collect.value:
            # extend current epoch
            if self.current_epoch is None:
                self.current_epoch = data.data
            else:
                self.current_epoch = np.concatenate([self.current_epoch, data.data], axis=1)

            if self.current_epoch.shape[1] >= (self.params.vamp.epoch_size.value + self.params.vamp.time_lag.value):
                # epoch is full, update model and start new epoch
                self.vamp_collect_changed(True)

        if self.model is None:
            return None

        # remove channel names, which are now components
        if "dim0" in data.meta["channels"]:
            del data.meta["channels"]["dim0"]

        # transform current data
        comps = self.model.transform(data.data.T).T
        return {"comps": (comps, data.meta)}

    def vamp_collect_changed(self, value):
        if self.current_epoch is not None:
            self.buffer.append(self.current_epoch.T)

            # update model
            tau = self.params.vamp.time_lag.value
            self.internal_model.partial_fit((self.current_epoch.T[:-tau], self.current_epoch.T[tau:]))
            self.model = self.internal_model.fetch_model()

            self.current_epoch = None

    def vamp_time_lag_changed(self, value):
        # refit a new model with updated time lag
        self.internal_model = self.make_model()
        self.internal_model.fit_from_timeseries(self.buffer)
        self.model = self.internal_model.fetch_model()

    def vamp_n_dims_changed(self, value):
        # refit a new model with updated number of dimensions
        self.internal_model = self.make_model()
        self.internal_model.fit_from_timeseries(self.buffer)
        self.model = self.internal_model.fetch_model()

    def make_model(self):
        from deeptime.decomposition import VAMP as VAMP_Model

        return VAMP_Model(lagtime=self.params.vamp.time_lag.value, dim=self.params.vamp.n_dims.value)



from os.path import join

import numpy as np

from goofi.data import Data, DataType
from goofi.node import Node
from goofi.params import BoolParam, FloatParam, IntParam, StringParam


class Classifier(Node):
    def config_input_slots():
        return {"data": DataType.ARRAY}

    def config_output_slots():
        return {
            "probs": DataType.ARRAY,
            "feature_importances": DataType.ARRAY,  # for RandomForest and LogisticRegression
        }

    def config_params():
        return {
            "classification": {
                "n_states": IntParam(2, 1, 10),
                "current_state": IntParam(1, 1, 10),
                "add_to_training": BoolParam(
                    False, doc="Add the incoming data to the training set, must be in the shape (n_features, n_samples)"
                ),
                "train": BoolParam(False, trigger=True, doc="Train the classifier"),
                "classifier_choice": StringParam(
                    "SVM", options=["NaiveBayes", "SVM", "RandomForest", "LogisticRegression", "KNeighbors"]
                ),
                "clear_training": BoolParam(False, trigger=True, doc="Clear the training set"),
                "loadname": StringParam("datafile"),
                "load_data": BoolParam(False, trigger=True),
                "savename": StringParam("savename"),
                "save_data": BoolParam(False, trigger=True),
            },
            "NaiveBayes": {"var_smoothing": FloatParam(1e-9, 1e-12, 1e-6)},
            "SVM": {
                "C": FloatParam(1.0, 0.1, 10.0, doc="Penalty parameter C of the error term"),
                "kernel": StringParam(
                    "rbf",
                    options=["linear", "poly", "rbf", "sigmoid"],
                    doc="Specifies the kernel type to be used in the algorithm",
                ),
                "gamma": StringParam(
                    "scale", options=["scale", "auto"], doc="Kernel coefficient for ‘rbf’, ‘poly’ and ‘sigmoid’"
                ),
            },
            "RandomForest": {
                "n_estimators": IntParam(100, 10, 1000, doc="Number of trees in the forest"),
                "max_depth": IntParam(3, 1, 100, doc="Maximum depth of the tree"),
                "min_samples_split": IntParam(10, 2, 100, doc="Minimum number of samples required to split a node"),
            },
            "LogisticRegression": {"C": FloatParam(1.0, 0.1, 10.0, doc="Inverse of regularization strength")},
            "KNeighbors": {
                "n_neighbors": IntParam(5, 1, 20, doc="Number of neighbors to use by default for kneighbors queries")
            },
        }

    def setup(self):
        from sklearn.ensemble import RandomForestClassifier
        from sklearn.linear_model import LogisticRegression
        from sklearn.naive_bayes import GaussianNB
        from sklearn.neighbors import KNeighborsClassifier
        from sklearn.svm import SVC

        self.RandomForestClassifier = RandomForestClassifier
        self.LogisticRegression = LogisticRegression
        self.GaussianNB = GaussianNB
        self.KNeighborsClassifier = KNeighborsClassifier
        self.SVC = SVC

        self.training_data = []
        self.training_labels = []
        self.classifier = None
        self.classifier_trained = False

    def process(self, data: Data):
        if data is None:
            return None
        self.loadname = self.params.classification["loadname"].value
        self.savename = self.params.classification["savename"].value
        self.loadpath = join(self.assets_path, self.loadname)
        self.savepath = join(self.assets_path, self.savename)

        if self.params.classification.add_to_training.value:
            transposed_data = data.data.T
            self.training_data.extend(transposed_data)
            self.training_labels.extend([self.params.classification.current_state.value] * transposed_data.shape[0])
            self.classifier = None
            self.classifier_trained = False
            # print("Added to training set.")  # Debug statement
            return None

        if self.params.classification.load_data.value:
            try:
                loaded_data = np.load(self.loadpath + ".npz")
                self.training_data = loaded_data["training_data"].tolist()
                self.training_labels = loaded_data["training_labels"].tolist()
                print(f"Loaded training data from {self.loadpath}.")  # Debug statement
            except Exception as e:
                print(f"Error during loading data: {e}")  # Debug statement
                return None

            # Automatically train when data is loaded
            self.params.classification.train.value = True

        if self.params.classification.save_data.value:
            try:
                np.savez(self.savepath, training_data=self.training_data, training_labels=self.training_labels)
                print(f"Saved training data to {self.savepath}.")  # Debug statement
            except Exception as e:
                print(f"Error during saving data: {e}")  # Debug statement
                return None

        if self.params.classification.clear_training.value:
            self.training_data = []
            self.training_labels = []
            self.classifier = None
            self.classifier_trained = False
            print("Training set cleared.")

        if self.params.classification.clear_training.value:
            self.training_data = []
            self.training_labels = []
            self.classifier = None
            self.classifier_trained = False
            print("Training set cleared.")

        if len(self.training_data) == 0:
            print("No training data.")
            return None

        if self.classifier is None:
            classifier_choice = self.params.classification.classifier_choice.value

            if classifier_choice == "NaiveBayes":
                self.classifier = self.GaussianNB(var_smoothing=self.params.NaiveBayes.var_smoothing.value)
            elif classifier_choice == "SVM":
                self.classifier = self.SVC(
                    C=self.params.SVM.C.value,
                    kernel=self.params.SVM.kernel.value,
                    gamma=self.params.SVM.gamma.value,
                    probability=True,
                )
            elif classifier_choice == "RandomForest":
                self.classifier = self.RandomForestClassifier(
                    n_estimators=self.params.RandomForest.n_estimators.value,
                    max_depth=self.params.RandomForest.max_depth.value,
                    min_samples_split=self.params.RandomForest.min_samples_split.value,
                )
            elif classifier_choice == "LogisticRegression":
                self.classifier = self.LogisticRegression(C=self.params.LogisticRegression.C.value)
            elif classifier_choice == "KNeighbors":
                self.classifier = self.KNeighborsClassifier(n_neighbors=self.params.KNeighbors.n_neighbors.value)

        if self.params.classification.train.value:
            # check if there are enough samples for each class
            for i in range(1, self.params.classification.n_states.value + 1):
                if self.training_labels.count(i) < 2:
                    print(f"Not enough samples for class {i} in training set.")
                    return None
            try:
                print(
                    f"Training data shape: {len(self.training_data)} samples, {len(self.training_data[0]) if self.training_data else 0} features per sample."
                )  # Debug statement
                self.classifier.fit(self.training_data, self.training_labels)
                self.classifier_trained = True
            except Exception as e:
                print(f"Error during fitting: {e}")  # Debug statement

        # check if the classifier has been trained
        if self.classifier_trained is False:
            print("Classifier not trained.")
            return None

        transposed_data = data.data.T
        probs = self.classifier.predict_proba(transposed_data)

        # After the classifier has been trained
        feature_importances = self.get_feature_importances()
        # if feature_importances is not None:
        # print("Feature Importances:", feature_importances)

        # create metadata including the classifier, the size of the training set for each class, and the number of features
        meta = {"classifier": self.params.classification.classifier_choice.value}
        for i in range(1, self.params.classification.n_states.value + 1):
            meta[f"training_set_size_{i}"] = self.training_labels.count(i)
        meta["n_features"] = len(self.training_data[0]) if self.training_data else 0
        meta_features = {}
        if "channels" in data.meta:
            meta_features["channels"] = data.meta["channels"]
        if "sfreq" in data.meta:
            meta["sfreq"] = data.meta["sfreq"]
            meta_features["sfreq"] = data.meta["sfreq"]
        return {"probs": (probs, meta), "feature_importances": (np.array(feature_importances), meta_features)}

    def get_feature_importances(self):
        """Retrieve feature importances from the classifier."""
        if isinstance(self.classifier, self.RandomForestClassifier):
            return self.classifier.feature_importances_
        elif isinstance(self.classifier, self.LogisticRegression):
            return self.classifier.coef_[0]
        elif isinstance(self.classifier, self.SVC) and self.params.SVM.kernel.value == "linear":
            return self.classifier.coef_[0]
        else:
            print("Feature importances not available for the current classifier or configuration.")
            return None
